import factory
import swapper
from django.contrib.contenttypes.models import ContentType

Reaction = swapper.load_model("baseapp_reactions", "Reaction")

# from baseapp_auth.tests.factories import UserFactory


class AbstractReactionFactory(factory.django.DjangoModelFactory):
    # TO DO: user = factory.SubFactory(UserFactory)
    reaction_type = factory.Faker("random_element", elements=Reaction.ReactionTypes)
    target_object_id = factory.SelfAttribute("target.id")
    target_content_type = factory.LazyAttribute(
        lambda o: ContentType.objects.get_for_model(o.target)
    )

    class Meta:
        exclude = ["target"]
        abstract = True



import pytest
from baseapp_comments.tests.factories import ClassroomCommentFactory

from .factories import CommentReactionFactory

pytestmark = pytest.mark.django_db


def test_student_can_see_reactions(django_user_student, graphql_user_student):
    user = django_user_student.user
    classroom = django_user_student.member.classroom
    comment = ClassroomCommentFactory(target=classroom, user=user)
    CommentReactionFactory(target=comment)
    response = graphql_user_student(
        "query { allClassrooms { totalCount edges { node { id commentsCount comments { edges { node { id content reactionsCount { total }}}} } } } }"
    )
    content = response.json()
    assert content["data"]["allClassrooms"]["totalCount"] == 1
    assert len(content["data"]["allClassrooms"]["edges"]) == 1
    assert (
        content["data"]["allClassrooms"]["edges"][0]["node"]["comments"]["edges"][0]["node"][
            "reactionsCount"
        ]["total"]
        == 1
    )


def test_instructor_can_see_reactions(django_user_instructor, graphql_user_instructor):
    user = django_user_instructor.user
    classroom = django_user_instructor.member.classroom
    comment = ClassroomCommentFactory(target=classroom, user=user)
    CommentReactionFactory(target=comment)
    response = graphql_user_instructor(
        "query { allClassrooms { totalCount edges { node { id commentsCount comments { edges { node { id content reactionsCount { total }}}} } } } }"
    )
    content = response.json()
    assert content["data"]["allClassrooms"]["totalCount"] == 1
    assert len(content["data"]["allClassrooms"]["edges"]) == 1
    assert (
        content["data"]["allClassrooms"]["edges"][0]["node"]["comments"]["edges"][0]["node"][
            "reactionsCount"
        ]["total"]
        == 1
    )



from graphene import relay

from .object_types import ReactionNode


class ReactionsQuery:
    # TO DO: fix permission, follow target until its not a Comment anymore and check if request.user has permission to see
    reaction = relay.Node.Field(ReactionNode)



import graphene
import swapper
from baseapp_core.graphql import RelayMutation, login_required
from baseapp_core.utils import get_content_type_by_natural_key
from django.contrib.contenttypes.models import ContentType
from django.utils.translation import ugettext_lazy as _
from graphql.error import GraphQLError
from graphql_relay import to_global_id
from graphql_relay.connection.arrayconnection import offset_to_cursor
from graphql_relay.node.node import from_global_id

from .object_types import ReactionNode, ReactionsNode, ReactionTypesEnum

Reaction = swapper.load_model("baseapp_reactions", "Reaction")


class ReactionToggle(RelayMutation):
    reaction = graphene.Field(ReactionNode._meta.connection.Edge, required=False)
    target = graphene.Field(ReactionsNode)
    reaction_deleted_id = graphene.ID(required=False)

    class Input:
        target_object_id = graphene.ID(required=True)
        target_content_type = graphene.String(
            required=False,
            description=_(
                "This overrides the GraphQL Type found in Relay's Global ID. Please use app.Model format."
            ),
        )
        reaction_type = graphene.Field(ReactionTypesEnum, required=True)

    @classmethod
    @login_required
    def mutate_and_get_payload(cls, root, info, **input):
        gid_type, gid = from_global_id(input.get("target_object_id"))
        object_type = info.schema.get_type(gid_type)
        target = object_type.graphene_type.get_node(info, gid)
        reaction_type = input["reaction_type"]

        target_content_type = input.get("target_content_type")
        if target_content_type:
            content_type = get_content_type_by_natural_key(target_content_type)
            target = content_type.get_object_for_this_type(pk=gid)
        else:
            content_type = ContentType.objects.get_for_model(target)

        # check https://github.com/queplanta/backend/blob/master/accounts/permissions.py#L9
        if "reaction-add" not in target.get_my_permissions(info.context):
            raise GraphQLError(
                _("You don't have permission to perform this action"),
                extensions={"code": "permission_required"},
            )
            # return cls(e)
        # target.has_permission(info.context, 'reaction')

        reaction, created = Reaction.objects.get_or_create(
            user=info.context.user,
            target_object_id=target.pk,
            target_content_type=content_type,
            defaults={"reaction_type": reaction_type},
        )
        if not created:
            if reaction.reaction_type == reaction_type:
                if "reaction-delete" not in target.get_my_permissions(info.context):
                    raise GraphQLError(
                        _("You don't have permission to perform this action"),
                        extensions={"code": "permission_required"},
                    )

                reaction_deleted_id = to_global_id(ReactionNode._meta.name, reaction.pk)
                reaction.delete()
                target.refresh_from_db()
                return ReactionToggle(target=target, reaction_deleted_id=reaction_deleted_id)

            if "reaction-change" not in target.get_my_permissions(info.context):
                raise GraphQLError(
                    _("You don't have permission to perform this action"),
                    extensions={"code": "permission_required"},
                )

            reaction.reaction_type = reaction_type
            reaction.save()

        target.refresh_from_db()

        return ReactionToggle(
            reaction=ReactionNode._meta.connection.Edge(node=reaction, cursor=offset_to_cursor(0)),
            target=target,
        )


class ReactionsMutations(object):
    reaction_toggle = ReactionToggle.Field()



import graphene
import swapper
from baseapp_core.graphql import CountedConnection
from django.contrib.contenttypes.models import ContentType
from graphene import relay
from graphene_django import DjangoObjectType
from graphene_django.filter import DjangoFilterConnectionField

Reaction = swapper.load_model("baseapp_reactions", "Reaction")

ReactionTypesEnum = graphene.Enum.from_enum(Reaction.ReactionTypes)


def create_object_type_from_enum(name, enum):
    fields = {}
    for reaction_type in enum:
        fields[reaction_type.name] = graphene.Int()
    fields["total"] = graphene.Int()
    return type(name, (graphene.ObjectType,), fields)


ReactionsCount = create_object_type_from_enum("ReactionsCount", Reaction.ReactionTypes)


class ReactionsNode(relay.Node):
    reactions_count = graphene.Field(ReactionsCount)
    reactions = DjangoFilterConnectionField(lambda: ReactionNode)
    my_reaction = graphene.Field(lambda: ReactionNode, required=False)

    def resolve_reactions(self, info, **kwargs):
        target_content_type = ContentType.objects.get_for_model(self)
        return Reaction.objects.filter(
            target_content_type=target_content_type,
            target_object_id=self.pk,
        ).order_by("-created")

    def resolve_my_reaction(self, info, **kwargs):
        if info.context.user.is_authenticated:
            target_content_type = ContentType.objects.get_for_model(self)
            return Reaction.objects.filter(
                target_content_type=target_content_type,
                target_object_id=self.pk,
                user=info.context.user,
            ).first()


class ReactionNode(DjangoObjectType):
    target = graphene.Field(relay.Node)
    reaction_type = graphene.Field(ReactionTypesEnum)

    class Meta:
        interfaces = (relay.Node,)
        model = Reaction
        fields = (
            "id",
            "user",
            "reaction_type",
            "created",
            "modified",
            "target",
        )
        filter_fields = {
            "id": ["exact"],
        }
        name = "Reaction"
        connection_class = CountedConnection



import baseapp_comments.tests.factories as f
import pytest
import swapper
from baseapp_comments.graphql.object_types import CommentNode
from graphql_relay import to_global_id

from ..models import ReactionTypes

Reaction = swapper.load_model("baseapp_reactions", "Reaction")

pytestmark = pytest.mark.django_db

REACTION_TOGGLE_GRAPHQL = """
    mutation ReactionToggleMutation($input: ReactionToggleInput!) {
        reactionToggle(input: $input) {
            reaction {
                node {
                    id
                    reactionType
                }
            }
        }
    }
"""


def test_anon_cant_react(graphql_client):
    comment = f.ClassroomCommentFactory()

    # TO DO: make to_global_id easier get, maybe like obj.relay_id
    comment_relay_id = to_global_id(CommentNode._meta.name, comment.pk)
    response = graphql_client(
        REACTION_TOGGLE_GRAPHQL,
        variables={"input": {"targetObjectId": comment_relay_id, "type": "LIKE"}},
    )
    content = response.json()
    assert content["errors"][0]["message"] == "authentication required"
    assert Reaction.objects.count() == 0


def test_user_can_react(graphql_user_client):
    comment = f.ClassroomCommentFactory()

    # TO DO: make to_global_id easier get, maybe like obj.relay_id
    comment_relay_id = to_global_id(CommentNode._meta.name, comment.pk)

    # create reaction with type LIKE
    graphql_user_client(
        REACTION_TOGGLE_GRAPHQL,
        variables={
            "input": {
                "targetObjectId": comment_relay_id,
                "type": ReactionTypes.LIKE.name,
            }
        },
    )
    assert Reaction.objects.count() == 1

    # change reaction with type LIKE to GRR
    graphql_user_client(
        REACTION_TOGGLE_GRAPHQL,
        variables={
            "input": {
                "targetObjectId": comment_relay_id,
                "type": ReactionTypes.GRR.name,
            }
        },
    )
    assert Reaction.objects.count() == 1
    assert Reaction.objects.filter(type=ReactionTypes.GRR).count() == 1

    # remove reaction
    graphql_user_client(
        REACTION_TOGGLE_GRAPHQL,
        variables={
            "input": {
                "targetObjectId": comment_relay_id,
                "type": ReactionTypes.GRR.name,
            }
        },
    )
    assert Reaction.objects.count() == 0
